#include <msp/fs/dir.h>
#include <msp/fs/stat.h>
#include <msp/strings/format.h>
#include "client.h"
#include "xinema.h"

using namespace std;
using namespace Msp;

Client::Client(Xinema &x, Net::StreamSocket *s):
	xinema(x),
	socket(s),
	stale(false)
{
	socket->signal_data_available.connect(sigc::mem_fun(this, &Client::data_available));
	socket->signal_end_of_file.connect(sigc::mem_fun(this, &Client::end_of_stream));

	xinema.signal_stream_created.connect(sigc::mem_fun(this, &Client::stream_created));
	XineStream *stream = xinema.get_stream();
	if(stream)
		stream_created(*stream);
}

void Client::data_available()
{
	char rbuf[1024];
	unsigned len;
	try
	{
		len = socket->read(rbuf, sizeof(rbuf));
	}
	catch(const std::exception &)
	{
		stale = true;
		return;
	}

	buffer.append(rbuf, len);

	string::size_type start = 0;
	while(1)
	{
		string::size_type newline = buffer.find('\n', start);
		if(newline==string::npos)
			break;

		try
		{
			process_command(buffer.substr(start, newline-start));
		}
		catch(const exception &e)
		{
			send_reply(string("error ")+e.what());
		}

		start = newline+1;
	}

	buffer.erase(0, start);
}

void Client::end_of_stream()
{
	stale = true;
}

XineStream &Client::get_stream() const
{
	XineStream *stream = xinema.get_stream();
	if(stream)
		return *stream;

	throw runtime_error("No stream");
}

void Client::process_command(const string &cmd)
{
	string::size_type space = cmd.find(' ');
	string keyword = cmd.substr(0, space);
	string args;
	if(space!=string::npos)
		args = cmd.substr(space+1);

	if(keyword=="list_directory")
		list_directory(args);
	else if(keyword=="play_file")
		xinema.play_file(args);
	else if(keyword=="play")
		get_stream().play();
	else if(keyword=="seek")
		get_stream().seek(lexical_cast<float>(args)*Time::sec);
	else if(keyword=="pause")
		get_stream().pause();
	else if(keyword=="stop")
		get_stream().stop();
	else
		throw runtime_error("Invalid command");
}

void Client::send_reply(const string &reply)
{
	Msp::MutexLock lock(mutex);
	try
	{
		socket->write(reply);
		socket->put('\n');
	}
	catch(const std::exception &)
	{
		stale = true;
	}
}

void Client::list_directory(const FS::Path &dn)
{
	list<string> files = FS::list_files(dn);

	send_reply("directory "+dn.str());
	for(list<string>::const_iterator i=files.begin(); i!=files.end(); ++i)
	{
		if(FS::is_dir(dn / *i))
			send_reply("subdir "+*i);
		else
			send_reply("file "+*i);
	}
}

void Client::stream_created(XineStream &stream)
{
	stream.signal_state_changed.connect(sigc::mem_fun(this, &Client::stream_state_changed));
	stream.signal_title_changed.connect(sigc::mem_fun(this, &Client::stream_title_changed));
	stream.signal_duration_changed.connect(sigc::mem_fun(this, &Client::stream_duration_changed));
	stream.signal_position_changed.connect(sigc::mem_fun(this, &Client::stream_position_changed));

	stream_state_changed(stream.get_state());

	string title = stream.get_title();
	if(!title.empty())
		send_reply("title "+title);

	if(const Time::TimeDelta &dur = stream.get_duration())
		stream_duration_changed(dur);
}

void Client::stream_state_changed(XineStream::State state)
{
	send_reply(format("state %s", state));
}

void Client::stream_title_changed(const string &title)
{
	send_reply("title "+title);
}

void Client::stream_duration_changed(const Time::TimeDelta &dur)
{
	send_reply(format("duration %.3f", dur/Time::sec));
}

void Client::stream_position_changed(const Time::TimeDelta &pos)
{
	if(abs(pos-last_position)>=Time::sec)
	{
		send_reply(format("position %.3f", pos/Time::sec));
		last_position = pos;
	}
}
